
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "base.h"
#include "convcar.h"
/*
*/
#include "errors.h"
#include "glifos.h"
#include "lista.h"
#include "imag.h"
#include "imagio.h"
#include "misc.h"
#include "opcns.h"
#include "pinta.h"
#include "tipos.h"

/************************ glifo **********************/

Boolean glifoCompuesto (tglifo glifo)
{
  /* depende de los tipos de glifo en glifo.h */
  return (glifo->tipo >= glifCompSec) &&
         (glifo->tipo <= glifCompTabla) ;

}

tglifo glifoNuevo ()
{
  tglifo nglifo;
  nglifo = (tglifo) malloc (sizeof(struct teglifo));
  nglifo->tipo        = glifNul ;
  nglifo->letras      = (char *) 0L;
  nglifo->venDeGlifo  = (tvent) 0L;
  nglifo->ven2DeGlifo  = (tvent) 0L;
  nglifo->glifBase    = (tglifo) 0L;
  nglifo->listaGlifos = (tlistadglifo) 0L;
  nglifo->base        = -1;       /* valor impropio */
  nglifo->alturax     = -1;
  nglifo->lizq        = -1;
  nglifo->lder        = -1;
  nglifo->recon       = reconNulo;
  nglifo->recindc     = 0;
  nglifo->recparecc   = 0;
  nglifo->parteLogAnt = (tglifo) 0L;
  nglifo->parteLogSig = (tglifo) 0L;
  nglifo->ideglifo    = 0;
  return nglifo;
}

tglifo duplicaGlifo (tglifo glifo)
{
  tglifo nglifo;
  nglifo = glifoNuevo ();
  nglifo->tipo       = glifo->tipo ;
  nglifo->letras     = strdup (glifo->letras);
  return nglifo;
}

tglifo glifoClinNuevo ()
{
  tglifo nglifo;
  nglifo = glifoNuevo ();
  nglifo->tipo       = glifClin ;
  return nglifo;
}

tglifo glifoBlanNuevo ()
{
  tglifo nglifo;
  nglifo = glifoNuevo ();
  nglifo->tipo       = glifBlan ;
  return nglifo;
}

tglifo glifoEspvNuevo ()
{
  tglifo nglifo;
  nglifo = glifoNuevo ();
  nglifo->tipo       = glifEspv ;
  return nglifo;
}

tglifo glifoLetNuevo (char *let)
{
  tglifo nglifo;
  nglifo = glifoNuevo ();
  nglifo->tipo       = glifCar ;
  nglifo->letras     = strdup (let);
  return nglifo;
}

void ponLetGlifo (tglifo glifo, char *let)
{
  free (glifo->letras);
  glifo->letras     = strdup (let);
}

tglifo glifoCompNuevo (int tipGlif)
{
  tglifo nglifo;
  nglifo = glifoNuevo ();
  nglifo->tipo        =  tipGlif ;
  nglifo->listaGlifos =  listadglifoNueva ();
  return nglifo;
}

tglifo glifoCompSecNuevo ()
{
  return glifoCompNuevo (glifCompSec);
}

tglifo glifoCompPalNuevo ()
{
  return glifoCompNuevo (glifCompPal);
}

tglifo glifoCompLinNuevo ()
{
  return glifoCompNuevo (glifCompLin);
}

tglifo glifoCompClmpNuevo ()
{
  return glifoCompNuevo (glifCompClmp);
}

tglifo glifoCompPaginaNuevo ()
{
  return glifoCompNuevo (glifCompPagina);
}

tglifo glifoCompPliegoNuevo ()
{
  return glifoCompNuevo (glifCompPliego);
}

tglifo glifoCompTablaNuevo ()
{
  return glifoCompNuevo (glifCompTabla);
}

tglifo glifoDeArrayDeGlifos (tglifo *glifos, int n)
{
  tglifo nglifo;
  int i;
  nglifo = glifoCompNuevo (glifCompPliego);
  for (i=0; i<n; i++)
    aniadeGlifoafinlista (nglifo->listaGlifos, glifos[i]);
  return nglifo;
}

void ponpadreGlifo (tglifo glifop, tglifo glifoh)
{
  glifoh->glifBase = glifop ;
}

void casaGlifoVent (tglifo glifo, tvent vent)
{
  vent->GLifoDeVen  = glifo ;
  glifo->venDeGlifo = vent ;
}

int separacionHorizEntreGlifos (tglifo glifo1, tglifo glifo2)
{
  int separacion;
  tvent vent1, vent2;
  if ((glifo1 == 0) || (glifo2 == 0))
    return -1;
  vent1 = glifo1->venDeGlifo;
  vent2 = glifo2->venDeGlifo;
  separacion = vent2->dx - (vent1->dx + vent1->tx);
  // describeGlifo (glifo2,1);
  return separacion;
}


void adoptaGlifo (tglifo glifop, tglifo glifoh)
{
  glifoh->glifBase = glifop ;
  aniadeGlifoafinlista (glifop->listaGlifos, glifoh);
}

tnodolglifo nodolglifoDeGlifo (tglifo glifo)
{
  tnodolglifo nodolGlifo;
  if (glifo->glifBase == 0L)
    return 0L;
  if (glifo->glifBase->listaGlifos == 0L)
    return 0L;
  nodolGlifo = glifo->glifBase->listaGlifos->primero;
  while (nodolGlifo !=  0L) {
    if (nodolGlifo->aglifo == glifo)
      break;
    if ( nodolGlifo->siguiente == 0L)
      return 0L;
    nodolGlifo = nodolGlifo->siguiente;
    }
  return nodolGlifo;
}

tglifo glifoBasenTipo (tglifo glifo, int tipoGlifo)
{
  tglifo glifoBasen;
  glifoBasen = glifo;
  if (glifoBasen == 0L) {
    printf ("error: ocre.glifos.glifoBasenTipo\n");
    exit (1);
    }
  while (glifoBasen->tipo != tipoGlifo) {
    glifoBasen = glifoBasen->glifBase ;
    if (glifoBasen == 0L) {
      printf ("error: ocre.glifos.glifoBasenTipo\n");
      exit (1);
      }
    }
  return glifoBasen;
}

tglifo glifoSiguienteDeGlifo (tglifo glifo)
{
  tnodolglifo nodolGlifo;
  nodolGlifo = nodolglifoDeGlifo (glifo);
  if (nodolGlifo == 0L)
    return 0L;
  if (nodolGlifo->siguiente == 0L)
    return 0L;
  else
    return nodolGlifo->siguiente->aglifo ;
}

tglifo glifoAnteriorDeGlifo (tglifo glifo)
{
  tnodolglifo nodolGlifo;
  nodolGlifo = nodolglifoDeGlifo (glifo);
  if (nodolGlifo == 0L)
    return 0L;
  if (nodolGlifo->anterior == 0L)
    return 0L;
  else
    return nodolGlifo->anterior->aglifo ;
}

/************************ lista glifos **********************/

tnodolglifo nodolglifoNuevo (tglifo glifo)
{
  tnodolglifo nnodolGlifo;
  nnodolGlifo = (tnodolglifo) malloc (sizeof(struct tenodolglifo));
  nnodolGlifo->aglifo = glifo;
  nnodolGlifo->siguiente = (tnodolglifo) 0L;
  nnodolGlifo->anterior  = (tnodolglifo) 0L;
  return nnodolGlifo ;
}

tlistadglifo listadglifoNueva ()
{
  tlistadglifo nlistadglifo;
  nlistadglifo = (tlistadglifo) malloc (sizeof(struct telistadglifo));
  nlistadglifo->primero = (tnodolglifo) 0L;
  nlistadglifo->ultimo  = (tnodolglifo) 0L;
  return nlistadglifo ;
}

Boolean listadglifoVacia (tlistadglifo ldglifo)
{
  return ldglifo->primero == (tnodolglifo) 0L ;
}

void aniadeGlifoacomlista (tlistadglifo ldglifo, tglifo glifo)
{
  tnodolglifo nnodolGlifo;
  nnodolGlifo = nodolglifoNuevo (glifo);
  if (ldglifo->primero == (tnodolglifo) 0L) {
    ldglifo->primero = nnodolGlifo ;
    ldglifo->ultimo  = nnodolGlifo ;
    }
  else {
    nnodolGlifo->siguiente  = ldglifo ->primero ;
    ldglifo->primero->anterior = nnodolGlifo ;
    ldglifo->primero = nnodolGlifo ;
    }
}

void aniadeGlifoafinlista (tlistadglifo ldglifo, tglifo glifo)
{
  tnodolglifo nnodolGlifo;
  nnodolGlifo = nodolglifoNuevo (glifo);
  if (ldglifo->primero == (tnodolglifo) 0L) {
    ldglifo->primero = nnodolGlifo ;
    ldglifo->ultimo  = nnodolGlifo ;
    }
  else {
    nnodolGlifo->anterior  = ldglifo ->ultimo ;
    ldglifo->ultimo->siguiente = nnodolGlifo ;
    ldglifo ->ultimo = nnodolGlifo ;
    }
}

void aniadeGlifoalistaAntes (tlistadglifo ldglifo, tglifo glifo, tnodolglifo nodolGlifo)
{
  if (ldglifo->primero == nodolGlifo)
    aniadeGlifoacomlista (ldglifo, glifo);
  else {
    tnodolglifo nnodolGlifo;
    nnodolGlifo = nodolglifoNuevo (glifo);
    nnodolGlifo->siguiente = nodolGlifo ;
    nnodolGlifo->anterior  = nodolGlifo->anterior;
    nodolGlifo->anterior->siguiente = nnodolGlifo;
    nodolGlifo->anterior            = nnodolGlifo;
  }
}

void eliminaNGlifodelista (tlistadglifo ldglifo, tnodolglifo nodolGlifo)
{
  tnodolglifo nodoantes, nododespues;
  if (ldglifo->primero == nodolGlifo)
    if (ldglifo->ultimo == nodolGlifo) {
      ldglifo->primero = 0L;
      ldglifo->ultimo  = 0L;
      }
    else {
      ldglifo->primero = ldglifo->primero->siguiente;
      ldglifo->primero->anterior = 0L;
      }
  else
    if (ldglifo->ultimo == nodolGlifo) {
      ldglifo->ultimo = ldglifo->ultimo->anterior;
      ldglifo->ultimo->siguiente = 0L;
      }
    else {
      nodoantes   = nodolGlifo->anterior;
      nododespues = nodolGlifo->siguiente;
      nodoantes->siguiente  = nododespues;
      nododespues->anterior = nodoantes;
      }
}

int longListagl (tlistadglifo ldglifo)
{
  int i=0;
  tnodolglifo nodolglifo;
  for (nodolglifo = ldglifo->primero; nodolglifo != 0L;nodolglifo = nodolglifo->siguiente) {
    i++ ;
    }
    return  i ;
}

void substituye2glifospor1 (tnodolglifo nodolGlifo, tglifo glifoNuevo)
{
  tlistadglifo listaGlifos;
  listaGlifos = nodolGlifo->aglifo->glifBase->listaGlifos;
  aniadeGlifoalistaAntes (listaGlifos, glifoNuevo, nodolGlifo);
  eliminaNGlifodelista (listaGlifos, nodolGlifo->siguiente);
  eliminaNGlifodelista (listaGlifos, nodolGlifo);
}

tglifo glifoUnionConSiguiente (tnodolglifo nodolGlifo)
{
  tglifo glifoNuevo;
  tvent ventanaNueva;
  unionGventanas ( &ventanaNueva, nodolGlifo->aglifo->venDeGlifo,
	                          nodolGlifo->siguiente->aglifo->venDeGlifo);
  glifoNuevo = glifoLetNuevo ("-?");
  casaGlifoVent (glifoNuevo, ventanaNueva);
  ponpadreGlifo (nodolGlifo->aglifo->glifBase, glifoNuevo);
  return glifoNuevo ;
}

Boolean enlazaLogGlifos (tglifo glif1, tglifo glif2)
{
  tglifo glifoBaseAnt, glifoBaseSig;
  if (glif1 == 0L) {
    if (glif2 == 0L)
      return FALSE;
    glifoBaseAnt = glifoAnteriorDeGlifo (glif1->glifBase);
    if (glifoBaseAnt == 0L)
      return FALSE;
    glif1 = glifoBaseAnt->listaGlifos->ultimo->aglifo;
    }
  if (glif2 == 0L) {
    glifoBaseSig = glifoSiguienteDeGlifo (glif1->glifBase);
    if ((glifoBaseSig == 0L) || !glifoCompuesto(glifoBaseSig))
      return FALSE;
    glif2 = glifoBaseSig->listaGlifos->primero->aglifo;
    }
  glif1->parteLogSig = glif2;
  glif2->parteLogAnt = glif1;
  return TRUE;
}

/************************ trazas ..  **********************/

static char * nombreTipoGlifo[] = {
  "glifCar", "glifBlan", "glifClin", "glifEspv", "glifNul", "glifArtf", "glifSecuen",
  "glifCompSec", "glifCompPal", "glifCompLin", "glifCompClmp",
  "glifCompPagina", "glifCompPliego", "glifCompTabla",
  "#"
};
static char * nombreTipoReconocimiento[] = {
  "reconNulo", "reconComp", "reconId3ft", "reconPreg", "reconMisc", "reconDicc",
  "#"
};

void describeGlifo (tglifo glifo, int niv)
{
  tnodolglifo nodolGlifo;
  tlistadglifo lisgli;
  printf ("g%d tipo: %s", niv, nombreTipoGlifo[glifo->tipo]);
  lisgli = glifo->listaGlifos ; 
  if (glifo->recon != 0){
    printf ("  reconocim: %s", nombreTipoReconocimiento[glifo->recon]);
    printf ("letras: %s", glifo->letras);
    }
  if (lisgli != 0L)
    printf ("  %3d - %3d subglifos\n", niv, longListagl(glifo->listaGlifos));
  else
    printf ("\n");
  if (glifo->venDeGlifo == 0L)
    printf ("g%d sin ventana asociada\n", niv);
  else {
    printf ("g%d ventana asociada:\n", niv);
    describeVentana (glifo->venDeGlifo);
    }
  printf (" base: %4d, alturax: %4d\n", glifo->base, glifo->alturax);
  if ((niv > 0) && (lisgli != 0L))
    for (nodolGlifo = lisgli->primero; nodolGlifo != 0L; nodolGlifo= nodolGlifo->siguiente)
      describeGlifo (nodolGlifo->aglifo, niv-1);
}

/************************ aplica recursivam. **********************/

// aplica recursivamente a un glifo de un tipo una funcion con unos datos
void aplicaGtRFncD (tglifo glifo, int tipo, void fn(tglifo, tpgen), tpgen datos)
{
  tnodolglifo nodolGlifo;
  if (glifo->tipo == tipo)
    fn (glifo, datos);
  else
    if (glifoCompuesto(glifo))
      for (nodolGlifo = glifo->listaGlifos->primero;
           nodolGlifo !=  0L;
           nodolGlifo = nodolGlifo->siguiente) {
        aplicaGtRFncD (nodolGlifo->aglifo, tipo, fn, datos);
        }
}

// aplica recursivamente a un glifo de un tipo una funcion (sin datos)
void aplicaGtRFn (tglifo glifo, int tipo, void fn(tglifo))
{
  tnodolglifo nodolGlifo;
  if (glifo->tipo == tipo)
    fn (glifo);
  else
    if (glifoCompuesto(glifo))
      for (nodolGlifo = glifo->listaGlifos->primero;
           nodolGlifo !=  0L;
           nodolGlifo = nodolGlifo->siguiente) {
        aplicaGtRFn (nodolGlifo->aglifo, tipo, fn);
        }
}

// ----------

void cuentaTiposenGlifoR0 (int *cntglifo, tglifo glifo)
{
  tnodolglifo nodolGlifo;
  cntglifo[glifo->tipo]++ ;
  if (glifoCompuesto(glifo))
    for (nodolGlifo = glifo->listaGlifos->primero;
         nodolGlifo !=  0L;
         nodolGlifo = nodolGlifo->siguiente) {
      cuentaTiposenGlifoR0 (cntglifo, nodolGlifo->aglifo);
      }
}

void cuentaTiposenGlifo (int **ncntglifo, tglifo glifo)
{
  int * cntglifo;
  int tg;
  cntglifo = (int *) malloc (glifZz * sizeof (int)); 
  for (tg = glifCar; tg <= glifZz; tg++)
    cntglifo [tg] = 0;
  cuentaTiposenGlifoR0 (cntglifo, glifo);
  *ncntglifo = cntglifo;
}

void imprimeRecuentoTiposenGlifo (int *cntglifo)
{
  int tg;
  for (tg = glifCar; tg < glifZz; tg++)
    printf ("%5d %s\n", cntglifo[tg], nombreTipoGlifo[tg]);
}

// ----------

void incrementaCnrecon (tglifo glifo, tpgen datos)
{
  int *cnrecon = (int *) datos;
  cnrecon[glifo->recon] ++ ;
}

void cuentaTiposReconocimientos (int **ncnrecon, tglifo glifo)
{
  int * cnrecon;
  int tr;
  cnrecon = (int *) malloc (reconZz * sizeof (int)); 
  // printf ("cuentaTiposReconocimientos\n");
  for (tr = reconNulo; tr <= reconZz; tr++)
    cnrecon [tr] = 0;
  aplicaGtRFncD (glifo, glifCar, incrementaCnrecon, cnrecon);
  *ncnrecon = cnrecon;
}

void imprimeRecuentoReconocimientos (int *cnrecon)
{
  int tr;
  for (tr = glifCar; tr < reconZz; tr++)
    printf ("%5d %s\n", cnrecon[tr], nombreTipoReconocimiento[tr]);
}

// ----------

void incrementaCnreconId3 (tglifo glifo, tpgen datos)
{
  int *cnrecon = (int *) datos;
  if (glifo->recon == reconId3ft) 
    cnrecon[glifo->recindc] ++ ;
}

void cuentaTiposReconocimientosId3 (int **ncnreconId3, tglifo glifo)
{
  int * cnreconId3;
  int i;
  cnreconId3 = (int *) malloc (5 * sizeof (int));   // cambiar con mas arboles
  // printf ("cuentaTiposReconocimientosId3\n");
  for (i = 0; i <= 5; i++)
    cnreconId3 [i] = 0;
  aplicaGtRFncD (glifo, glifCar, incrementaCnreconId3, cnreconId3);
  *ncnreconId3 = cnreconId3;
}

void imprimeRecuentoReconocimientosId3 (int *cnreconId3)
{
  int i;
  for (i = 0; i <= 5; i++)
    if (cnreconId3[i] != 0)
      printf ("%5d %2d\n", cnreconId3[i], i);
}

void imprimeTasaReconocimientosId3 (int *cnreconId3, int *cntglifo)
{
  double tasa;
  int i;
  int k;
  k = 0;
  for (i = 0; i <= 5; i++)
    k += cnreconId3[i] * i;
  tasa = (k * 100.0) / (cntglifo[glifCar] * 5);
  printf ("tasa: %7.3f\n\n", tasa);
}

// ----------

int nmaxcompc = 25;     // cambiar si no es suficiente (ROLLO)

void incrementaCnreconComp (tglifo glifo, tpgen datos)
{
  int *cnrecon = (int *) datos;
  if (glifo->recon == reconComp) 
    cnrecon[glifo->recparecc] ++ ;
}

void cuentaTiposReconocimientosComp (int **ncnreconComp, tglifo glifo)
{
  int * cnreconComp;
  int i;
  cnreconComp = (int *) malloc (nmaxcompc * sizeof (int));
  // printf ("cuentaTiposReconocimientosComp\n");
  for (i = 0; i <= nmaxcompc; i++)
    cnreconComp [i] = 0;
  aplicaGtRFncD (glifo, glifCar, incrementaCnreconComp, cnreconComp);
  *ncnreconComp = cnreconComp;
}

void imprimeRecuentoReconocimientosComp (int *cnreconComp)
{
  int i;
  for (i = 0; i <= nmaxcompc; i++)
    if (cnreconComp[i] != 0)
      printf ("%5d %2d\n", cnreconComp[i], i);
}

void imprimeTasaReconocimientosComp (int *cnreconComp, int *cntglifo)
{
  double tasa;
  int i;
  int k;
  k = 0;
  for (i = 0; i <= nmaxcompc; i++)
    k += cnreconComp[i] * i;
  tasa = (k * 100.0) / (cntglifo[glifCar] * 5);
  printf ("tasa: %7.3f\n\n", tasa);
}

/************************ escribe letras ..  **********************/

void escribePalabrasdeGlifosenListaR0 (tlistadglifo ldglifo)
{
  tnodolglifo nodolGlifo;
  char carsus[1024];
  nodolGlifo = ldglifo->primero;
  while (nodolGlifo !=  0L) {
    switch (nodolGlifo->aglifo->tipo) {
      case glifCompSec :
        escribePalabrasdeGlifosenListaR0 (nodolGlifo->aglifo->listaGlifos);
        break ;
      case glifCar :
	ssalidaDeUtf (carsus, nodolGlifo->aglifo->letras);
        printf ("%s", carsus);
        break ;
      }
    nodolGlifo = nodolGlifo->siguiente;
    }
}

void escribePalabradeGlifo (tpgen glifo)
{
  escribePalabrasdeGlifosenListaR0( ((tglifo)glifo)->listaGlifos );
  printf ("\n");
}

void escribePalabradeGlifoL (tpgen glifo)
{
  tglifo gli = (tglifo) glifo;
  escribePalabrasdeGlifosenListaR0(gli->listaGlifos );
  if (gli->parteLogSig != 0L)
    escribePalabrasdeGlifosenListaR0(gli->parteLogSig->listaGlifos );
  printf ("\n");
}

void escribePalabradeGlifoyPares (tpgen glifo)
{
  printf ("_ %2d ", parejasDeLetrasEnPalabra(glifo));
  escribePalabrasdeGlifosenListaR0( ((tglifo)glifo)->listaGlifos );
  printf ("\n");
}

void escribePalabrasdeGlifosenListaR (tlistadglifo ldglifo, int *n, Boolean log)
{
  tnodolglifo nodolGlifo;
  tglifo glifo;
  int np;
  char carsus[1024];
  np = *n;
  nodolGlifo = ldglifo->primero;
  while (nodolGlifo !=  0L) {
    glifo = nodolGlifo->aglifo ;
    switch (glifo->tipo) {
      case glifCompSec :
        escribePalabrasdeGlifosenListaR (glifo->listaGlifos, &np, log);
        break ;
      case glifCar :
	ssalidaDeUtf (carsus, glifo->letras);
        printf ("%s", carsus);
        break ;
      case glifClin :
        if (!log) {
          printf ("\n");
	  }
        break ;
      case glifBlan :
      case glifEspv :
        break ;
      case glifCompPliego :
      case glifCompPagina :
      case glifCompClmp :
      case glifCompLin :
        escribePalabrasdeGlifosenListaR (glifo->listaGlifos, &np, log);
        break ;
      case glifCompPal :
        np++ ;
        if (log) {
	  if (glifo->parteLogAnt == 0L ) {
            printf ("\n");
	    escribePalabrasdeGlifosenListaR (glifo->listaGlifos, &np, log);
	    }
	  if (glifo->parteLogSig != 0L ) {
	    escribePalabrasdeGlifosenListaR (glifo->parteLogSig->listaGlifos, &np, log);
	    }
          }
	else {
          printf ("\n");
          printf ("%4d ", np);
	  {
	  /*
	  int i;
	  for (i=0; i<6; i++)
            printf ("%3d ",glifo->venDeGlifo->caracIncl[i]);
	    */
	  }
	  printf ("%3d ",glifo->venDeGlifo->incl);
	  printf ("_ %2d ", parejasDeLetrasEnPalabra(glifo));
	  escribePalabrasdeGlifosenListaR (glifo->listaGlifos, &np, log);
	  }
        break ;
      }
    nodolGlifo = nodolGlifo->siguiente;
    }
  *n = np;
}

void escribePalabrasdeGlifosenLista (tlistadglifo ldglifo)
{
  int np;
  np = 0;
  printf ("\n%5d ", np);
  escribePalabrasdeGlifosenListaR (ldglifo, &np, FALSE);
  printf ("\n");
}

void escribePalabrasLogdeGlifosenLista (tlistadglifo ldglifo)
{
  int np;
  np = 0;
  escribePalabrasdeGlifosenListaR (ldglifo, &np, TRUE);
  printf ("\n");
}

void escribeGlifoBlanco (tglifo glifo)
{
  int sep;
  tglifo glifoAnt, glifoSig;
  glifoAnt = glifoAnteriorDeGlifo (glifo);
  glifoSig = glifoSiguienteDeGlifo (glifo);
  //
  if ((glifoAnt->tipo) == glifCompPal)
    glifoAnt = glifoAnt->listaGlifos->ultimo->aglifo;
  if ((glifoSig->tipo) == glifCompPal)
    glifoSig = glifoSig->listaGlifos->primero->aglifo;
  //
  sep = separacionHorizEntreGlifos (glifoAnt, glifoSig);
  if (sep < 16)
    printf ("%s","");
  else if (sep < 30)
    printf (" ");
  else
    printf ("   ");
}

void escribeLetrasdeGlifosenListaR (tlistadglifo ldglifo, int *n)
{
  tnodolglifo nodolGlifo;
  tglifo glifo;
  char carsus[1024];
  int nc;
  nc = *n;
  // if (nc == 575) tirarPrograma ();
  nodolGlifo = ldglifo->primero;
  while (nodolGlifo !=  0L) {
    glifo = nodolGlifo->aglifo;
    switch (glifo->tipo) {
      case glifCar :
//        printf (" %4d %4d ", glifo->venDeGlifo->dx, glifo->venDeGlifo->dx+glifo->venDeGlifo->tx-1);
        if (u_opcion)
	  strcpy (carsus, glifo->letras);
        else
	  ssalidaDeUtf (carsus, glifo->letras);
        printf ("%s", carsus);
        nc++ ;
        break ;
      case glifEspv :
      case glifClin :
        if (R_opcion) {
          printf ("\r");
	  }
        printf ("\n");
        if (N_opcion) {
          printf ("%5d ", nc);
          }
        break ;
      case glifBlan :
        printf (" ");
        // escribeGlifoBlanco (glifo);
        break ;
      case glifCompLin :
	/*   ====
	printf ("%6d\n", glifo->linAlinizq);
	*/
	if (glifo->linAlinizq >= 10)
	  printf(" ");
      case glifCompPliego :
      case glifCompPagina :
      case glifCompClmp :
      case glifCompPal :
      case glifCompSec :
        escribeLetrasdeGlifosenListaR (glifo->listaGlifos, &nc);
        break ;
      }
    nodolGlifo = nodolGlifo->siguiente;
    }
  *n = nc;
}

void escribeLetrasdeGlifosenLista (tlistadglifo ldglifo, int * nc)
{
  if (N_opcion) {
    printf ("%5d ", *nc);
    }
  escribeLetrasdeGlifosenListaR (ldglifo, nc);
  if (N_opcion) {
    if (R_opcion) {
      printf ("\r");
      }
    printf ("\n");
    }
}

void escribeEnFdLetrasdeGlifosenLista (FILE *fd, tlistadglifo ldglifo)
{
  char carsus[1024];
  tnodolglifo nodolGlifo;
  nodolGlifo = ldglifo->primero;
  while (nodolGlifo !=  0L) {
    switch (nodolGlifo->aglifo->tipo) {
      case glifCar :
	ssalidaDeUtf (carsus, nodolGlifo->aglifo->letras);
        fprintf (fd, "%s", carsus);
        break ;
      case glifEspv :
      case glifClin :
        if (R_opcion) {
          fprintf (fd, "\r");
          }
        fprintf (fd, "\n");
        break ;
      case glifBlan :
        fprintf (fd, " ");
        break ;
      case glifCompPliego :
      case glifCompPagina :
      case glifCompClmp :
      case glifCompLin :
      case glifCompPal :
      case glifCompSec :
        escribeEnFdLetrasdeGlifosenLista (fd, nodolGlifo->aglifo->listaGlifos);
        break ;
      }
    nodolGlifo = nodolGlifo->siguiente;
    }
}

void escribeEnFnomLetrasdeGlifosenLista (char *nf, tlistadglifo ldglifo)
{
  FILE * fd;
  fd = fopen (nf, "w");
  compruebaFd (fd, nf, "glifos.escribeEnFnomLetrasdeGlifosenLista");
  escribeEnFdLetrasdeGlifosenLista (fd, ldglifo);
  fclose (fd);
}

void ponNombreSim (char *nf, char *ndir, char *lc, int nc)
{
  char nomu[128];
  sustrDeUtf8 (nomu, lc);
  sprintf (nf, "%s/%06d-%s.pgm", ndir, nc, nomu);
}

void ponNombreSim2 (char *nf, char *ndir, char *apndir, char *lc, int nc)
{
  char nomu[128];
  sustrDeUtf8 (nomu, lc);
  sprintf (nf, "%s%s/%06d-%s.pgm", ndir, apndir, nc, nomu);
}

/************************ alin. base y altx  **********************/

void baseyalturaxDeGlifo (int *base, int *alturax, tglifo glifo)
{
  int baseEnLinea, alturaxEnLinea;
  tglifo gl;
  gl = glifo;
  while (gl->base == -1) {
    if (gl->glifBase == 0L) {
      printf ("ocre: error en glifos.baseyalturaxDeGlifo, no encontrada base\n");
      tirarPrograma ();
      exit (2);
      }
    gl = gl->glifBase ;
    /*
    if (gl->base != 0L)
      gl = gl->glifBase ;
    else
      gl = gl->venDeGlifo->venBase->glifoDeVen;
      */
    }
  baseEnLinea    = gl->base ;
  alturaxEnLinea = gl->alturax ;
  *base     = baseEnLinea      + gl->venDeGlifo->dy;
  *alturax  = alturaxEnLinea   + gl->venDeGlifo->dy;
}

void alineamientoNormalizado (int *dyNorm, int *dy2Norm, int *tyNorm, tglifo glifo)
{
  int base, alturax;
  int baseMenosAlturax;
  int dy, ty, dy2;
  int dxr;
  baseyalturaxDeGlifo (&base, &alturax, glifo) ;
  // dy = glifo->venDeGlifo->dy;
  ty = glifo->venDeGlifo->ty;
  desplazamientosBase2 (&dxr, &dy, glifo->venDeGlifo);
  dy2 = dy + ty;
  baseMenosAlturax = max(1, base -alturax);
  *dyNorm  = 112+ ((dy  - alturax) * 32 + 16)/ baseMenosAlturax;
  *dy2Norm = 112+ ((dy2 - alturax) * 32 + 16)/ baseMenosAlturax;
  *tyNorm  = *dy2Norm - *dyNorm ;
}

char ** comentarioAlineamiento (tglifo glifo)
                /* e inclinacion si es <> 0 */
{
  char buf[128];
  char ** comen2;
  int dyNorm, tyNorm, dy2Norm;
  alineamientoNormalizado ( &dyNorm, &dy2Norm, &tyNorm, glifo);
  // dyNorm = 96; dy2Norm = 144; tyNorm = 48;   //  SOLO para numeros de pagina
  if (glifo->venDeGlifo->incl == 0) 
    sprintf (buf, "dyNorm=%4d dy2Norm=%4d tyNorm=%4d", dyNorm, dy2Norm, tyNorm);
  else
    sprintf (buf, "dyNorm=%4d dy2Norm=%4d tyNorm=%4d inclvc=%4d",
                   dyNorm, dy2Norm, tyNorm, glifo->venDeGlifo->incl);
  listaAarrayConcopia (&comen2, 1, buf);
  return comen2;
}

/************************ aplicadoR al glifo **********************/
// ----------

typedef struct tenmdirn{
  char * ndir;
  int  * n;
  } * tndirn;

// Boolean soloReconPreg = TRUE;
// Boolean soloReconPreg = FALSE;
// Boolean escribeTodos  = TRUE;
 Boolean escribeTodos  = FALSE;

// void kkk (tglifo glifo, int nc)
// {
//   printf ("\n");
//   printf ("%3d\n",nc);
//   describeGlifo (glifo, 2);
//   printf ("#\n");
//   describeGlifo (glifo->glifBase, 0);
//   describeGlifo (glifo->glifBase->glifBase, 0);
//   describeGlifo (glifo->glifBase->glifBase->glifBase, 0);
// }

void escribeVentanadeGlifoCar (tglifo glifo, tpgen gndirn)
{
  char nf[128];
  char ** comen2;
  int nc;
  tndirn ndirn;
  // if ( (! soloReconPreg) ||
  //      (glifo->recon == reconPreg) ) {
  ndirn = (tndirn) gndirn;
  nc = *(ndirn->n);
  // if ((nc>=64) && (nc <=70))
  //  kkk (glifo, nc);
  if (escribeTodos ||
      ! ((glifo->recon == reconId3ft) &&
         (glifo->recindc >=4) )) {
    ponNombreSim (nf, ndirn->ndir, glifo->letras, nc);
    comen2 = comentarioAlineamiento (glifo);
    escribeVentanaLimMascomentarios (glifo->venDeGlifo, nf, comen2);
    free (comen2);
  }
  else {
    ponNombreSim2 (nf, ndirn->ndir, "b", glifo->letras, nc);
    comen2 = comentarioAlineamiento (glifo);
    escribeVentanaLimMascomentarios (glifo->venDeGlifo, nf, comen2);
    free (comen2);
    }
  nc ++ ;
  *(ndirn->n) = nc;
}

void escribeVentanasdeGlifosenLista (tglifo glifo, int *nc, char *ndir)
{
  struct tenmdirn endirn;
  endirn.n    = nc;
  endirn.ndir = ndir;
  aplicaGtRFncD (glifo, glifCar, escribeVentanadeGlifoCar, (tpgen) &endirn);
}

// ---------- en obras-1

void numeraGlifoCar (tglifo glifo, tpgen gndirn)
{
  int nc;
  tndirn ndirn;
  ndirn = (tndirn) gndirn;
  nc = *(ndirn->n);
  glifo->ideglifo = nc;
  nc ++ ;
  *(ndirn->n) = nc;
}

void numeraGlifosenLista (tglifo glifo, int *nc)
{
  struct tenmdirn endirn;
  endirn.n    = nc;
  aplicaGtRFncD (glifo, glifCar, numeraGlifoCar, (tpgen) &endirn);
}


// ----------

void marcaGlifoCarNoReconocido (tglifo glifo)
{
  if (glifo->recon == reconNulo)
    pintaRectIntVent (glifo->venDeGlifo);
}

void marcaGlifosNoReconocidos (tglifo glifoColumnaP) 
{
  aplicaGtRFn (glifoColumnaP, glifCar, marcaGlifoCarNoReconocido);
}

// ----------

void anyadeGlifoAlista(tglifo glifo, tpgen datos)
{
  aniadeafinlista ((lista) datos, (tpgen) glifo);
}

void anyadeGlifoLAlista(tglifo glifo, tpgen datos)
{
  if (glifo->parteLogAnt == 0L)
    aniadeafinlista ((lista) datos, (tpgen) glifo);
}

// palabras N
void listadePalabrasEnGlifo (lista *nlista, tglifo glifo)
{
  lista l;
  l = listaNueva ();
  aplicaGtRFncD (glifo, glifCompPal, anyadeGlifoAlista, (tpgen) l);
  *nlista = l;
}

// palabras Logicas
void listadePalabrasLenGlifo (lista *nlista, tglifo glifo)
{
  lista l;
  l = listaNueva ();
  aplicaGtRFncD (glifo, glifCompPal, anyadeGlifoLAlista, (tpgen) l);
  *nlista = l;
}

// ----------

tglifo glifoUltimoconchar (tglifo glifoc)
{
  tglifo gul;
  tnodolglifo nodolGlifo;
  switch (glifoc->tipo) {
    case glifCompLin :
    case glifCompPal :
    case glifCompSec :
      for (nodolGlifo = glifoc->listaGlifos->ultimo;
           nodolGlifo != 0L; nodolGlifo = nodolGlifo->anterior){
	gul = glifoUltimoconchar (nodolGlifo->aglifo) ;
	if (gul != 0L)
          return gul;
	}
      return 0L;
      break ;
    case glifCar :
      // if (glifoc->recon != reconNulo)
      return glifoc ;
      // else
      //   return 0L;
      break ;
    default :
      return 0L;
      break ;
    }
}

Boolean glifoAcabaenchar (tglifo glifoc, char ch)
{
  tglifo glifoul;
  int l;
  glifoul = glifoUltimoconchar (glifoc) ;
  if (glifoul == 0L)
    return FALSE;
  if (glifoul->recon == reconNulo)
    return FALSE;
  l = strlen (glifoul->letras);
  return glifoul->letras[l-1] == ch ;
}

// ----------

void anyadeaParteLetrasConocidas (tglifo glifo, tpgen datos)
{
  char *tira;
  lista lis = (lista) datos;
  if (glifo->recon == reconNulo){
    //// anyadeafinlista parte nueva con tira vacia
    tira = (char *) malloc (100);
    tira[0] = '\0';
    aniadeafinlista (lis, tira);
    }
  else {
    //// anyade al ultimo elemento los chars del glifo
    //   como utf8
    strncat ((char *)(lis->ultimo->aitem), glifo->letras, 100);
    //   como ascii
    // char carsus[1024];
    // ssalidaDeUtf (carsus, glifo->letras);
    // strncat ((char *)(lis->ultimo->aitem), carsus, 100);
    }
}

lista partesDePalabraNenGlifo (tglifo glifo)
{
  lista lista;
  char *tira;
  lista = listaNueva ();
  //// anyadeafinlista parte nueva con tira vacia
    tira = (char *) malloc (100);
    tira[0] = '\0';
    aniadeafinlista (lista, tira);
  aplicaGtRFncD (glifo, glifCar, anyadeaParteLetrasConocidas, (tpgen) lista);
  return lista;
}

// 1 caracter de longitud 1 octeto
void quitaAlfinal (char *tira, char car)
{
  int l;
  l = strlen (tira);
  if (l != 0)
    if (tira[l-1] == car)
      tira[l-1] = '\0';
}

lista partesDePalabraLenGlifo (tglifo glifo)
{
  lista lista;
  char *tira;
  lista = listaNueva ();
  //// anyadeafinlista parte nueva con tira vacia
    tira = (char *) malloc (100);
    tira[0] = '\0';
    aniadeafinlista (lista, tira);
  aplicaGtRFncD (glifo, glifCar, anyadeaParteLetrasConocidas, (tpgen) lista);
  if (glifo->parteLogSig == 0L)
    return lista;
  else {
    // '-' es un caracter de logitud un octeto (en utf8)
    quitaAlfinal ( (char *)lista->ultimo->aitem, '-');
    aplicaGtRFncD (glifo->parteLogSig, glifCar, anyadeaParteLetrasConocidas, (tpgen) lista);
    return lista;
    }
}

// ----------

void cuentaGlifosIncognita(tglifo glifo, tpgen datos)
{
  int * n = (int *) datos;
  if (glifo->recon == reconNulo)
    (* n) ++ ;
}

int nincognitasEnGlifo (tglifo glifo)
{
  int n = 0;
  aplicaGtRFncD (glifo, glifCar, cuentaGlifosIncognita, (tpgen) &n);
  return n;
}

int nincognitasEnLGlifo (tglifo glifo)
{
  int n = 0;
  aplicaGtRFncD (glifo, glifCar, cuentaGlifosIncognita, (tpgen) &n);
  if (glifo->parteLogSig == 0L)
    return n;
  else {
    aplicaGtRFncD (glifo->parteLogSig, glifCar, cuentaGlifosIncognita, (tpgen) &n);
    return n;
    }
}

// ----------

void anyadealistaGlifoIncognita(tglifo glifo, tpgen datos)
{
  if (glifo->recon == reconNulo)
    aniadeafinlista ((lista) datos, (tpgen) glifo);
}

lista listaIncognitasEnGlifo (tglifo glifo)
{
  lista lista;
  lista = listaNueva ();
  aplicaGtRFncD (glifo, glifCar, anyadealistaGlifoIncognita, (tpgen) lista);
  return lista ;
}

lista listaIncognitasEnGlifoL (tglifo glifo)
{
  lista lista;
  lista = listaNueva ();
  aplicaGtRFncD (glifo, glifCar, anyadealistaGlifoIncognita, (tpgen) lista);
  if (glifo->parteLogSig == 0L)
    return lista ;
  else {
    aplicaGtRFncD (glifo->parteLogSig, glifCar, anyadealistaGlifoIncognita, (tpgen) lista);
    return lista ;
    }
}

// ----------

typedef struct telistayletr{
  lista lista;
  char  * letra;
  } * tlistayletr;

void anyadealistaGlifoCarConLetras(tglifo glifo, tpgen gdatos)
{
  tlistayletr datos = (tlistayletr) gdatos;
  if (strcmp(glifo->letras, datos->letra)==0)
    aniadeafinlista (datos->lista, (tpgen) glifo);
}

lista listaGlifosConLetraEnGlifo (tglifo glifo, char *letra)
{
  struct telistayletr datos;
  datos.lista = listaNueva ();
  datos.letra = letra;
  aplicaGtRFncD (glifo, glifCar, anyadealistaGlifoCarConLetras, (tpgen) &datos);
  return datos.lista ;
}

// ----------

typedef struct te2char{
  char ch1, ch2;
} * t2char;

typedef struct te2str {
  char *chs1, *chs2;
} * t2str;

void cambiaEnglifoCar(tglifo glifo, tpgen datos)
{
  char cant  = ((t2char)datos)->ch1;
  char cdesp = ((t2char)datos)->ch2;
  if (glifo->letras[0] == cant)
    glifo->letras[0] = cdesp;
}

void cambiaNEnGlifo (tglifo glifo, char cant, char cdesp)
{
  struct te2char datos;
  datos.ch1 = cant;
  datos.ch2 = cdesp;
  aplicaGtRFncD (glifo, glifCar, cambiaEnglifoCar, (tpgen) &datos);
}

void cambiaEnglifoStr(tglifo glifo, tpgen datos)
{
  char *sant  = ((t2str)datos)->chs1;
  char *sdesp = ((t2str)datos)->chs2;
  if (strcmp(glifo->letras, sant) == 0)
    glifo->letras = sdesp;
}

void cambiaStrNEnGlifo (tglifo glifo, char *sant, char *sdesp)
{
  struct te2str datos;
  datos.chs1 = sant;
  datos.chs2 = sdesp;
  aplicaGtRFncD (glifo, glifCar, cambiaEnglifoCar, (tpgen) &datos);
}

// ----------

void enlazaPalabrasPartidas (tglifo glifoColumnaP)
{
  tglifo glifoLin, glifoPalult;
  tnodolglifo nodolGlifo;
  for (nodolGlifo = glifoColumnaP->listaGlifos->primero;
       nodolGlifo != 0L; nodolGlifo = nodolGlifo->siguiente) {
    glifoLin = nodolGlifo->aglifo;
    if (glifoLin->tipo == glifCompLin) {
      glifoPalult = glifoLin->listaGlifos->ultimo->anterior->aglifo ; // -camb.linea
      // describeGlifo (glifoPalult, 2);
      if (glifoAcabaenchar (glifoPalult, '-')) {
        // enlaza ultimaPalabra esa linea con siguiente palabra (0L).
        enlazaLogGlifos (glifoPalult, (tglifo) 0L);
        }
      }
    }
}


void enlazaPalabraPartida (tglifo glifoPalabra)
{
  tglifo glifoLinea;
  glifoLinea = glifoPalabra->glifBase;
  // printf (">-->");
  // describeGlifo (glifoLinea, 2);
  // printf (">->");
  // describeGlifo (glifoPalabra, 2);
  if ((glifoPalabra->tipo == glifCompPal) &&
      (glifoPalabra->glifBase->listaGlifos->ultimo->anterior->aglifo
                                              // -camb.linea
         == glifoPalabra)) {
    if (glifoAcabaenchar (glifoPalabra, '-')) {
      // enlaza ultimaPalabra esa linea con siguiente palabra (0L).
      enlazaLogGlifos (glifoPalabra, (tglifo) 0L);
      }
    }
}

